;;; sml-hilite.el 
;;; 08-Nov-94, Fritz Knabe <knabe@ecrc.de>
;;;
;;; Highlighting for sml-mode using hilit19. Takes advantage of
;;; hilit19's ability to install arbitrary functions to do parsing
;;; (rather than just regexps) to get correct highlighting of strings
;;; and comments.
;;;
;;; Put this code *after* the (require 'hilit19) in your .emacs.
;;; Alternatively, put it in an (eval-after-load "hilit19" ...).
;;;
;;; The highlighting package does not currently appear to belong to
;;; XEmacs -- they favour `font-lock'. Shame, but font-lock patterns
;;; may be added here later.

(require 'hilit19)
(provide 'sml-hilite)

(cond ((and (x-display-color-p) (eq hilit-background-mode 'light))
       ;; for SML
       (hilit-translate sml-comment	'firebrick-italic)
       (hilit-translate sml-string	'ForestGreen-italic)
       (hilit-translate sml-keyword	'blue-bold))
      ((and (x-display-color-p) (eq hilit-background-mode 'dark))
       ;; for SML
       (hilit-translate sml-comment	'moccasin-italic)
       (hilit-translate sml-string	'green-italic)
       (hilit-translate sml-keyword	'cyan-bold))
      (t
       ;; for SML
       (hilit-translate sml-comment	'default-italic)
       (hilit-translate sml-string	'default-bold-italic)
       (hilit-translate sml-keyword	'default-bold)))

(hilit-set-mode-patterns
 'sml-mode
 '((kn-hilit-sml-string-find "" sml-string)
   (kn-hilit-sml-comment-find "" sml-comment)
   ;; The old patterns
   ;;("\"" "[^\\]\"" sml-string)
   ;;("(\\*" "\\*)" sml-comment)
   ("\\(\\`\\|[^_']\\)\\<\\(abstraction\\|abstype\\|and\\|andalso\\|as\\|\
case\\|datatype\\|else\\|end\\|eqtype\\|exception\\|do\\|fn\\|fun\\|\
functor\\|handle\\|if\\|in\\|infix\\|infixr\\|let\\|local\\|nonfix\\|of\\|\
op\\|open\\|overload\\|raise\\|rec\\|sharing\\|sig\\|signature\\|struct\\|\
structure\\|then\\|type\\|val\\|while\\|with\\|withtype\\|orelse\\)\\>\
\\(\\'\\|[^_']\\)" 2 sml-keyword)))

(defun kn-hilit-sml-string-find (dummy)
  "Find an SML string and return (START . END); if none, returns nil. 
Skips over potentially nested comments when searching for the start of the
string. Skips over \f...f\ (where f is whitespace) sequences in strings."
  (let ((nest 0)
	(continue t)
	st en)
    (while (and continue
		(re-search-forward "\\(\"\\)\\|\\((\\*\\)\\|\\(\\*)\\)" nil t))
      (cond
       ((match-beginning 1) (setq continue (> nest 0)))
       ((match-beginning 2) (setq nest (+ nest 1)))
       ((match-beginning 3) (setq nest (- nest 1)))))
    (if (not continue)
	(progn
	  (setq st (match-beginning 1))
	  (while (and (re-search-forward
		       "\\(\"\\)\\|\\(\\\\\\s-*\\\\\\)\\|\\(\\\\\"\\)" nil t)
		      (cond
		       ((match-beginning 1) (setq en (point)) nil)
		       ((match-beginning 2) t)
		       ((match-beginning 3) t))))
	  (and en (cons st en))))))

(defun kn-hilit-sml-comment-find (dummy)
  "Find an SML comment and return (START . END); if none, returns nil.
Handles nested comments. Ensures that the comment starts outside of a string."
  (let ((continue t)
	(nest 1)
	st en)
    (while (and continue
		(re-search-forward "\\(\"\\)\\|\\((\\*\\)" nil t))
      (cond
       ((match-beginning 1)
	(while (and (re-search-forward
		     "\\(\"\\)\\|\\(\\\\\\s-*\\\\\\)\\|\\(\\\\\"\\)" nil t)
		    (cond
		     ((match-beginning 1) nil)
		     ((match-beginning 2) t)
		     ((match-beginning 3) t)))))
       ((match-beginning 2) (setq continue nil))))
    (if (not continue)
	(progn
	  (setq st (match-beginning 2))
	  (setq continue t)
	  (while (and continue
		      (re-search-forward "\\((\\*\\)\\|\\(\\*)\\)" nil t))
	    (cond
	     ((match-beginning 1) (setq nest (+ nest 1)))
	     ((match-beginning 2)
	      (setq nest (- nest 1)) (setq continue (> nest 0)))))
	  (if (not continue)
	      (cons st (match-end 2)))))))
