;;; sml-menus.el
;;; Simple menus for sml-mode

(require 'easymenu) 

;;; That's Per Abrahamsen's easymenu, not FSF's. (At least if you want
;;; the menus to work for emacsen other than FSF's. Tested with FSF
;;; 19.26 and 19.27, and XEmacs (Lucid) 19.11.)

;;; Menus appear only when the insertion point is in an sml-mode
;;; buffer. They should appear automatically as long as sml-mode can
;;; find this file and easymenu.el. Otherwise they won't!

;;; If you load sml-proc.el to run an inferior ML process -- or even a
;;; superior one, who knows? -- the menu will be updated with a few
;;; more handy commands you'r fingers may have forgotten.

;;; Added sml-sml-nj-093 choice in the menu, and in sml-proc.el. H.Leiss
;;; Removed sml-sml-nj-093 choice in the menu
;;; Automata-Lib in the menu and in sml-proc.el.      H.Leiss, 6.3.2000

(provide 'sml-menus)

;; WARNING! "Process" menu gets modified when process code is loaded.

(defconst sml-menu
  (list "SML"
	(list "Process" ["Load ML process commands" (require 'sml-proc) t]) 
	["electric pipe"     sml-electric-pipe t]
	["insert SML form"   sml-insert-form t]
	(list "Forms" 
	      ["abstype"     (sml-abstype) t]
              ["datatype"    (sml-datatype) t]
              ["-" nil nil]
              ["let"         (sml-let) t]
              ["local"       (sml-local) t]
              ["case"        (sml-case) t]
              ["--" nil nil]
              ["signature"   (sml-signature) t]
              ["functor"     (sml-functor) t]
              ["structure"   (sml-structure) t]
              ["---" nil nil]
              ["abstraction" (sml-abstraction) t])
        (list "Format/Mode Variables"
              ["indent region"    sml-indent-region t]
              ["outdent"          sml-back-to-outer-indent t]
              ["set indent-level" sml-indent-level t]
              ["set pipe-indent"  sml-pipe-indent t]
              ["toggle type-of-indent" (sml-type-of-indent) t]
              ["toggle nested-if-indent"(sml-nested-if-indent) t]
              ["toggle case-indent" (sml-case-indent) t]
              ["toggle electric-semi-mode" (sml-electric-semi-mode) t])
        ["-----" nil nil]
	["Help on Emacs SML-mode"      describe-mode t]
	["SML mode version"  sml-mode-version t]
	))

(easy-menu-define sml-mode-menu
    sml-mode-map
    "Menu used in sml-mode."
    sml-menu)

;; Make's sure they appear in the menu bar when sml-mode-map is active.

(add-hook 'sml-mode-hook 
           (function (lambda () (easy-menu-add sml-mode-menu sml-mode-map))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Update the sml-menu (idea borrowed from Per Abrahamsen's LaTeX mode)
;;; (You got to credit where yer ideas come from -- nothing's original.)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ALL these functions are defined in sml-proc.el you won't get sense
;;; out of them before commands in sml-proc.el are loaded.

(defconst sml-proc-menu
  (list ["load ML source file"  sml-load-file t]
	["-" nil nil]
	["send buffer contents" sml-send-buffer t]
	["send region"          sml-send-region t]
	["send paragraph"       sml-send-function t]
	["goto next error"      sml-next-error t]
	["--" nil nil]
	["run Standard ML of New Jersey 110 " sml-sml-nj t]
	["HTML Basis-Library of SML-NJ 110" sml-sml-nj-BasisLibrary t]
	["HTML Automata-Library for SML-NJ 110" sml-AutomataLibrary t]
;;;	["Standard ML of New Jersey 0.93" sml-sml-nj-093 t]
;;;	["Help on New Jersey ML 0.93 (info)" sml-sml-nj-info t]
))

;;; "Process" must be defined on sml-menu already -- if it in't these
;;; won't get added! (Does nothing if menus aren't defined/required.)

(defun sml-menu-update ()
  (if (not (and (featurep 'sml-menus) (equal mode-name "SML")))
      ()
    (unwind-protect
	(easy-menu-change '("SML") "Process" sml-proc-menu))))


;;; When this file is loaded... (sml-proc.el says something similar)

(if (not (featurep 'sml-proc))
    ()
  (add-hook 'activate-menubar-hook 'sml-menu-update))

